/**
 * An implementation of an undirected graph using an adjacency list 
 * Vertex labels are strings 
 * (c) 2021 Aaron Bauer
 */

import java.util.HashMap;
import java.util.HashSet;
import java.util.Map;
import java.util.Set;

public class UndirectedGraph {
    private Map<String, Set<String>> adjLists;

    // construct an empty graph
    public UndirectedGraph() {
        adjLists = new HashMap<>();
    }
    
    // throw an exception if v is not present in the graph
    private void validateVertex(String v) {
        if (!adjLists.containsKey(v)) {
            throw new IllegalArgumentException("Vertex " + v + " has not been added to the graph");
        }
    }

    // return a new Graph that is a copy of this graph
    public UndirectedGraph copy() {
        UndirectedGraph g = new UndirectedGraph();
        g.adjLists = new HashMap<>();
        for (Map.Entry<String, Set<String>> adj : adjLists.entrySet()) {
            g.adjLists.put(adj.getKey(), new HashSet<>());
            for (String v : adj.getValue()) {
                g.adjLists.get(adj.getKey()).add(v);
            }
        }
        return g;
    }

    // add vertex v to the graph
    public void addVertex(String v) {
        adjLists.put(v, adjLists.getOrDefault(v, new HashSet<String>()));
    }

    // add an undirected edge from u to v to the graph
    public void addEdge(String u, String v) {
        if (u.equals(v)) throw new IllegalArgumentException("self edges not supported");
        addVertex(u);
        addVertex(v);
        adjLists.get(u).add(v);
        adjLists.get(v).add(u);
    }

    // return a set of the vertices in the graph
    public Set<String> V() {
        return new HashSet<>(adjLists.keySet());
    }

    // return an iterable of the vertices neighboring v
    public Iterable<String> neighbors(String v) {
        validateVertex(v);
        return adjLists.get(v);
    }

    // return the degree of vertex v
    public int degree(String v) {
        validateVertex(v);
        return adjLists.get(v).size();
    }

    // remove vertex v, if present
    public void removeVertex(String v) {
        if (containsVertex(v)) {
            for (String u : neighbors(v)) {
                adjLists.get(u).remove(v);
            }
            adjLists.remove(v);
        }
    }

    // remove the undirected edge from u to v from the graph
    public void removeEdge(String u, String v) {
        validateVertex(u);
        validateVertex(v);
        adjLists.get(u).remove(v);
        adjLists.get(v).remove(u);
    }

    // return true if the graph contains vertex v
    public boolean containsVertex(String v) {
        return adjLists.containsKey(v);
    }

    // return true if the graph contains an edge from u to v
    public boolean containsEdge(String u, String v) {
        return containsVertex(u) && containsVertex(v) && adjLists.get(u).contains(v);
    }

    // return the number of vertices in the graph
	public int numVertices() {
		return adjLists.size();
	}

    // return the number of edges in the graph
	public int numEdges() {
        int count = 0;
        for (Set<String> s : adjLists.values()) {
            count += s.size();
        }
        // each undirected edge will count twice, once for each direction
        return count / 2;
	}

    /**
     * Returns a string representation of this graph.
     *
     * @return the number of vertices <em>V</em>, followed by the number of edges
     *         <em>E</em>, followed by the <em>V</em> adjacency lists
     */
    @Override
    public String toString() {
        StringBuilder s = new StringBuilder();
        s.append(numVertices() + " vertices, " + numEdges() + " edges\n");
        for (Map.Entry<String, Set<String>> entry : adjLists.entrySet()) {
            s.append(entry.getKey() + ": ");
            for (String w : entry.getValue()) {
                s.append(w + " ");
            }
            s.append("\n");
        }
        return s.toString();
    }
}